#!/usr/bin/env python3

import math
import numpy as np
from numba import jit, prange
from scipy import integrate
from datetime import datetime as dt
import statistics

# Notes: Using SI units here
### ------------------------------ Defining all unchanging parameters / constants to be used ------------------------------ ###
pi = math.pi
mu0 = 4*pi*1e-7                                                                          # Permeability of free space (in H/m)         
inc_chi = 3 + 9.035e-6                                                                   # Difference in magnetic susceptibility between bead and carrying medium
eta = 1e-3                                                                               # Dynamic viscosity of fluid (in Pa s, assume water at 20degC for now)
T = 293                                                                                  # Temperature (in K)
kB = 1.38065e-23                                                                         # Boltzmann constant (in J/K)
###############################################################################################################################

### ------------------------------ Maximum current prediction function based on wire radii ------------------------------ ###
wire_radii = np.array([0, 100, 130, 160, 205, 255, 320, 405, 510, 645, 815, 1025, 1295])
currents = np.array([0, 0.3, 0.5, 0.7, 1, 2, 3, 5, 7, 10, 20, 30, 50])

# Assume linear relationship between provided data points (X, y)
@jit(nopython=True)
def predict_current(rw, ndp):
    left_index, right_index = 0, 1
    # Scan the provided data to identify where the input argument rw lies in the wire_array
    while right_index <= len(wire_radii) - 1:
        # When found, perform linear intrapolation between data points
        if rw >= wire_radii[left_index] and rw <= wire_radii[right_index]:
            pred_current = currents[left_index] + (currents[right_index] - currents[left_index])*((rw - wire_radii[left_index])/(wire_radii[right_index] - wire_radii[left_index]))
            #print(f"Found input argument current at between index {left_index} and {right_index}.")
            break
        else:
            left_index += 1
            right_index += 1

    if right_index == len(wire_radii):
        raise Exception("Input argument not in model data range!")
    else:
        # Round predicted current to number of decimal places specified by ndp
        factor = 10**ndp
        rounded_current = math.floor(pred_current*factor)/factor
        return rounded_current
###############################################################################################################################

### ------------------------------ Initialize all parameters that remain unchanged in one particle track ------------------------------ ###
@jit(nopython=True)
def init_params(particle_radius, flowrate, sep_length, tubing_radius, kappa, rcap_ratio, power_status):
    """
    This function initializes all parameters needed to start particle tracking (except particle size) in SI units
    The output is a numpy array (0: Particle radius, 1: Pressure drop, 2: Separator length, 3: Wire radius, 4: Tubing radius, 5: Capture radius, 6: Current in wire, 7: Kappa, 8: Mean residence time, 9: Stokes-Einstein relation)
    The arguments to this init_params function are as follows:
    particle_radius = int/float nm
    flowrate = float ml/min
    sep_length = float mm
    tubing_radius = float um
    kappa = *dimensionless*
    rcap_ratio = *dimensionless* -> rcap_ratio = ((rcap-rw)/(rc-rw))
    power_status = *boolean* (denotes if current is flowing or not)
    """
    # Initializing particle radius
    rp = particle_radius*1e-9                               # in m

    # Calculating pressure drop
    Qdot_m3s = flowrate/6e7
    deltap_perlength = Qdot_m3s/( (pi*((tubing_radius*1e-6)**4)*(1 - kappa**2)/(8*eta)) * (((1 - kappa**4)/(1 - kappa**2)) - ((1 - kappa**2)/(math.log(1/kappa)))) )
    deltap = deltap_perlength*sep_length*1e-3               # in Pa

    # Initializing seperator length
    ls = sep_length*1e-3                                    # in m

    # Initializing tubing radius
    rc = tubing_radius*1e-6                                 # in m

    # Calculating wire radius
    rw = kappa*rc                                           # in m

    # Calculating mean residence time
    annular_volume = pi*((rc**2) - (rw**2)) * ls            # in m3
    mean_residence_time = annular_volume/Qdot_m3s           # in s

    # Calculating capture radius
    rcap = (rc - rw)*rcap_ratio + rw                        # in m

    # Predict current for this wire radius
    if power_status == True:
        current = predict_current(rw=rw*1e6, ndp=2)                 # in A
    else:
        current = 0

    # Initializing Brownian parameter (Stokes-Einstein equation)
    D_coeff = kB*T/(6*pi*eta*(particle_radius*1e-9))

    # Prepare and return all unchanging parameters in one particle trajectory tracking
    all_params = np.array([rp, deltap, ls, rw, rc, rcap, current, kappa, mean_residence_time, D_coeff])
    return all_params
####################################################################################################################################################

### ------------------------------ PARTICLE TRACKING ALGORITHM (WITH TRAJECTORY OUTPUT) ------------------------------ ###
@jit(nopython=True)                 # '(float64[:], float64, float64, boolean)'
def track_particle_fully(all_params, start_r, time_des, Brownian, seed):
    """
    This function outputs a particle trajectory (from the start of the separator, i.e. z = 0) for a tube-in-a-tube separator for MNPs, with z-direction fluid flow.

    Input arguments are as follows:
    particle_radius = int/float type (in nm)
    all_params = *generated from init_params* list type
    start_r = int/float type (in um)
    time_des = float type (in s)
    Brownian = bool type
    """
    # Initialize random seed
    np.random.seed(int(seed))

    # Initialize particle radius
    rp = all_params[0]                  # in m

    # Initialize all other parameters
    p_drop = all_params[1]              # in Pa
    sep_length = all_params[2]          # in m
    wire_radius = all_params[3]         # in m
    sep_radius = all_params[4]          # in m
    capture_radius = all_params[5]      # in m
    I = all_params[6]                   # in Pa
    kappa = all_params[7]               # [-]
    tau = all_params[8]                 # in s
    D_coeff = all_params[9]
                                                                
    # Initializing pos_r (calculated from center of wire) and pos_z
    pos_r = start_r*1e-6
    pos_z = 0

    # Initializing time discretization and total time counter
    inc_t = time_des                                           
    total_time = 0                                                      # Total time taken

    # Initialize residence time multiplier
    tau_multiplier = 10

    # Pre-allocate time_array, pos_r_array, pos_z_array, based on the number on the fact that they will not exceed 10*mean_residence_time
    time_array = np.arange(0, math.floor(tau_multiplier*tau/inc_t)+1, inc_t)
    pos_r_array = np.zeros(math.floor(tau_multiplier*tau/inc_t)+1)
    pos_r_array[0] = pos_r*1e6                                          # In units of um
    pos_z_array = np.zeros(math.floor(tau_multiplier*tau/inc_t)+1)
    pos_z_array[0] = pos_z*1e3                                          # In units of mm
    
    # Initialize wire_touch_cnt
    wire_touch_cnt = 0

    # Initialize time_step_cnt
    time_step_cnt = 1

    # Track particle movement (with fluid flow)
    while True:
        # Calculate fluid velocity and in turn particle velocity in the r and z directions
        vf_z = (p_drop*(sep_radius**2)/(4*eta*sep_length)) * (1 - ((pos_r/sep_radius)**2) - ((1 - kappa**2)/math.log(1/kappa)) * math.log(sep_radius/pos_r))
        vp_r = (((rp**2)*(I**2)*(mu0)*(inc_chi))/(18*(pi**2)*eta)) * (-1/(pos_r ** 3))
        vp_z = vf_z
        # Calculate new pos_r and pos_z (if statement to evaluate with or without Brownian motion)
        if Brownian == False:
            pos_r_new = pos_r + vp_r*inc_t
            pos_z_new = pos_z + vp_z*inc_t
        else:
            lBr = np.random.normal(loc=0, scale=math.sqrt(2*D_coeff*inc_t))
            lBz = np.random.normal(loc=0, scale=math.sqrt(2*D_coeff*inc_t))
            pos_r_new = pos_r + vp_r*inc_t + lBr
            pos_z_new = pos_z + vp_z*inc_t + lBz

        # Check if the pos_r_new exceeds that allowable by separator radius, or touches the wall of the separator
        if pos_r_new + rp >= sep_radius:
            pos_r_new = pos_r + vp_r*inc_t - lBr    # This ensures the particle moves away from the wall

        # Check if pos_r_new will touch the wire surface in the next time step
        if pos_r_new - rp <= wire_radius:
            pos_r_new = pos_r                       # Particle will not have its radial position updated if it is projected to touch wire surface in the next time step
            wire_touch_cnt += 1                     # Increase wire_touch_cnt by 1

        # Check if the particle has exceeded 10 times the average residence time or has reached the end of the separator
        # Remember, pos_r, pos_z, rp, and wire_radius are all in SI units (i.e. m)
        term_cond_1 = bool(total_time >= tau_multiplier*tau)        # When trajectories terminate with term_cond_1, the total_time recorded would be just one step beyond the 10*mean_residence_time. Correspondingly, the final positions recorded would have been one step beyond 10*mean_residence_time
        term_cond_2 = bool(pos_z_new + rp >= sep_length)            # When trajectories terminate with term_cond_2, the total_time recorded would be one time step just before particle leaves channel. Correspondingly, the final positions recorded would be just one time step before particle leaves channel
        term_cond_3 = bool(wire_touch_cnt > 100)                    # If a particle has touched the wire 100 times, terminate particle trajectory.
        if term_cond_1 or term_cond_2 or term_cond_3:               # There is a small possibility that all 3 termination conditions are satisfied
            break
        else:
            # Change next element in position arrays to new r and z positions
            pos_r_array[time_step_cnt] = pos_r_new*1e6
            pos_z_array[time_step_cnt] = pos_z_new*1e3

            # Reinitialize pos_r and pos_z with new values
            pos_r = pos_r_new
            pos_z = pos_z_new

            # Append updated total_time counter
            total_time += inc_t
            #time_array[time_step_cnt] = total_time
            
            # Update time_step_cnt
            time_step_cnt += 1

    # Trim zeros from back of time_array, pos_r_array, pos_z_array
    time_array = time_array[:time_step_cnt] #np.trim_zeros(time_array, 'b')
    pos_r_array = pos_r_array[:time_step_cnt] #np.trim_zeros(pos_r_array, 'b')
    pos_z_array = pos_z_array[:time_step_cnt] #np.trim_zeros(pos_z_array, 'b')

    # Assign a capture status to this trajectory
    capture_status = bool(pos_r_array[-1]*1e-6 + rp <= capture_radius)

    # Output particle radius, starting radial coordinates, and time taken
    # print(f"Particle radius: {particle_radius} nm. Starting radial coordinate: {start_r} um. Time taken: {total_time} s. TermCond1: {term_cond_1}. TermCond2: {term_cond_2}. TermCond3: {term_cond_3}\n")

    return time_array, pos_r_array, pos_z_array, total_time, capture_status

###############################################################################################################################################################

### ------------------------------ PARTICLE TRACKING ALGORITHM (WITHOUT TRAJECTORY OUTPUT) ------------------------------ ###
@jit(nopython=True)                 # '(float64[:], float64, float64, boolean)'
def track_particle(all_params, start_r, time_des, Brownian, seed):
    """
    This function outputs the capture status of a single trajectory.

    Input arguments are as follows:
    particle_radius = int/float type (in nm)
    all_params = *generated from init_params* list type
    start_r = int/float type (in um)
    time_des = float type (in s)
    Brownian = bool type
    """
    # Initialize random seed
    np.random.seed(int(seed))

    # Initialize particle radius
    rp = all_params[0]                  # in m

    # Initialize all other parameters
    p_drop = all_params[1]              # in Pa
    sep_length = all_params[2]          # in m
    wire_radius = all_params[3]         # in m
    sep_radius = all_params[4]          # in m
    capture_radius = all_params[5]      # in m
    I = all_params[6]                   # in Pa
    kappa = all_params[7]               # [-]
    tau = all_params[8]                 # in s
    D_coeff = all_params[9]
                                                                
    # Initializing pos_r (calculated from center of wire) and pos_z
    pos_r = start_r*1e-6
    pos_z = 0

    # Initializing time discretization and total time counter
    inc_t = time_des                                           
    total_time = 0                                                      # Total time taken

    # Initialize residence time multiplier
    tau_multiplier = 10
    
    # Initialize wire_touch_cnt
    wire_touch_cnt = 0

    # Initialize time_step_cnt
    time_step_cnt = 1

    # Track particle movement (with fluid flow)
    while True:
        # Calculate fluid velocity and in turn particle velocity in the r and z directions
        vf_z = (p_drop*(sep_radius**2)/(4*eta*sep_length)) * (1 - ((pos_r/sep_radius)**2) - ((1 - kappa**2)/math.log(1/kappa)) * math.log(sep_radius/pos_r))
        vp_r = (((rp**2)*(I**2)*(mu0)*(inc_chi))/(18*(pi**2)*eta)) * (-1/(pos_r ** 3))
        vp_z = vf_z
        # Calculate new pos_r and pos_z (if statement to evaluate with or without Brownian motion)
        if Brownian == False:
            pos_r_new = pos_r + vp_r*inc_t
            pos_z_new = pos_z + vp_z*inc_t
        else:
            lBr = np.random.normal(loc=0, scale=math.sqrt(2*D_coeff*inc_t))
            lBz = np.random.normal(loc=0, scale=math.sqrt(2*D_coeff*inc_t))
            pos_r_new = pos_r + vp_r*inc_t + lBr
            pos_z_new = pos_z + vp_z*inc_t + lBz

        # Check if the pos_r_new exceeds that allowable by separator radius, or touches the wall of the separator
        if pos_r_new + rp >= sep_radius:
            pos_r_new = pos_r + vp_r*inc_t - lBr    # This ensures the particle moves away from the wall

        # Check if pos_r_new will touch the wire surface in the next time step
        if pos_r_new - rp <= wire_radius:
            pos_r_new = pos_r                       # Particle will not have its radial position updated if it is projected to touch wire surface in the next time step
            wire_touch_cnt += 1                     # Increase wire_touch_cnt by 1

        # Check if the particle has exceeded 10 times the average residence time or has reached the end of the separator
        # Remember, pos_r, pos_z, rp, and wire_radius are all in SI units (i.e. m)
        term_cond_1 = bool(total_time >= tau_multiplier*tau)        # When trajectories terminate with term_cond_1, the total_time recorded would be just one step beyond the 10*mean_residence_time. Correspondingly, the final positions recorded would have been one step beyond 10*mean_residence_time
        term_cond_2 = bool(pos_z_new + rp >= sep_length)            # When trajectories terminate with term_cond_2, the total_time recorded would be one time step just before particle leaves channel. Correspondingly, the final positions recorded would be just one time step before particle leaves channel
        term_cond_3 = bool(wire_touch_cnt > 1000)                    # If a particle has touched the wire 1000 times, terminate particle trajectory.
        if term_cond_1 or term_cond_2 or term_cond_3:               # There is a small possibility that all 3 termination conditions are satisfied
            break
        else:
            # Reinitialize pos_r and pos_z with new values
            pos_r = pos_r_new
            pos_z = pos_z_new
            # Append updated total_time counter
            total_time += inc_t
            # Update time_step_cnt
            time_step_cnt += 1

    # Assign a capture status to this trajectory
    capture_status = bool(pos_r + rp <= capture_radius)

    return capture_status

# #np.random.seed(42)
# begin_time1 = dt.now()
# all_params = init_params(particle_radius=100, flowrate=0.1, sep_length=500, tubing_radius=500, kappa=0.7, rcap_ratio=0.5)
# # Run 1
# capture_status = track_particle(all_params=all_params, start_r=420, time_des=0.01, Brownian=True)
# print(capture_status, f"{dt.now() - begin_time1}")
# begin_time2 = dt.now()
# # Run 2
# capture_status = track_particle(all_params=all_params, start_r=420, time_des=0.01, Brownian=True)
# print(capture_status, f"{dt.now() - begin_time2}")
##################################################################################################################################################################

### ------------------------------ GENERATE ARRAY OF STARTING RADIAL POSITIONS ------------------------------ ###
def generate_startrpos(all_params, resolution, extra_info, n_sims, seed=42):
    """
    This function calculates the cummulative distribution function (numerically) for ANNULAR flow and can output an array of starting radial positions based upon the CDF.

    Input arguments are as follows:
    all_params = numpy array type   (generated from init_params function)
    resolution = int type           (determines the resolution of the cdf)
    extra_info = bool type          (determines if cdf, pdf, norm_factor, resolution pos r array, and MRCE is to be produced)
    n_sims = int type               (generates n_sims number of starting radial position)
    
    """
    # Initialize random seed
    np.random.seed(seed)

    # Initialize particle radius and all other parameters (in SI units)
    rp = all_params[0]                  # in m
    p_drop = all_params[1]              # in Pa
    sep_length = all_params[2]          # in m
    wire_radius = all_params[3]         # in m
    sep_radius = all_params[4]          # in m
    capture_radius = all_params[5]      # in m
    kappa = all_params[7]               # [-]

    # Prepare upper and lower limits of integration, as well as other terms of integration
    rpos_uplim = sep_radius - rp
    rpos_downlim = wire_radius + rp

    r_power1 = lambda r_up, r_down : r_up - r_down                  # Definitive integral of 1 wrt r (i.e. [r](upbound, lowbound))
    r_power3 = lambda r_up, r_down : (r_up**3) - (r_down**3)        # Definitive integral of r**2 wrt r (i.e. [r**3](upbound, lowbound))
    r_intgr_lnr = lambda r_up, r_down : (r_up*math.log(r_up) - r_up) - (r_down*math.log(r_down) - r_down)

    term1 = r_power1(rpos_uplim, rpos_downlim)
    term2 = -(1/(3*(sep_radius**2))) * r_power3(rpos_uplim, rpos_downlim)
    term3 = - (((1 - kappa**2)*math.log(sep_radius))/math.log(1/kappa)) * r_power1(rpos_uplim, rpos_downlim)
    term4 = ((1 - kappa**2)/math.log(1/kappa)) * r_intgr_lnr(rpos_uplim, rpos_downlim)

    # Analytical integration of fluid velocity profile over the annular radius
    norm_factor_k = (p_drop*(sep_radius**2)/(4*eta*sep_length)) * (term1 + term2 + term3 + term4)

    # Normalize the fluid velocity profile by norm_factor_k to obtain the probability distribution function (pdf_array)
    #print(type(rpos_downlim), type(rpos_uplim), type(resolution))
    # pos_r_list = np.array([rpos_downlim])
    # pos_r_cnt = 1
    # increment = (rpos_uplim - rpos_downlim)/(resolution-1)
    # while pos_r_cnt <= resolution:
    #     last_index = pos_r_cnt - 1
    #     if pos_r_cnt == resolution:
    #         pos_r_list = np.append(pos_r_list, [rpos_uplim])
    #     else:
    #         pos_r_list = np.append(pos_r_list, [pos_r_list[last_index] + increment])

    pos_r_np = np.linspace(start=rpos_downlim, stop=rpos_uplim, num=resolution)
    pdf_array = np.array([])
    vfz_array = np.array([])
    for i in range(len(pos_r_np)):
        pos_r = pos_r_np[i]
        vf_z = (p_drop*(sep_radius**2)/(4*eta*sep_length)) * (1 - ((pos_r/sep_radius)**2) - ((1 - kappa**2)/math.log(1/kappa)) * math.log(sep_radius/pos_r))
        pd = vf_z/norm_factor_k
        pdf_array = np.append(pdf_array, [pd])
        vfz_array = np.append(vfz_array, [vf_z])

    # Obtain the cummulative probability distribution of the starting radial position of the particle based upon annular flow
    cdf_array = integrate.cumtrapz(pdf_array, pos_r_np, initial=0)
    average_velocity = statistics.mean(vfz_array)               # in m/s

    # Calculate minimum required capture efficiency (MRCE) for the given annular flow profile/distribution function and the specified capture radius
    # Perform a binary search algorithm to return the indices that will bound capture radius. Result is an array pos_r_list[lhs:rhs] that will contain value of capture radius
    lhs, rhs = 0, len(pos_r_np) - 1
    while rhs - lhs > 2:
        target_index = math.floor((lhs + rhs)/2)
        if pos_r_np[target_index] >= capture_radius:
            rhs = target_index + 1                  # Retain the >0.5 value
        else:
            lhs = target_index
    # Interpolate to obtain MRCE
    gradient = (cdf_array[rhs] - cdf_array[lhs])/(pos_r_np[rhs] - pos_r_np[lhs])
    constant = cdf_array[lhs] - gradient*pos_r_np[lhs]
    MRCE = (gradient*capture_radius + constant) * 100

    # Generate an array of starting radial positions using the CDF
    start_rpos_array = np.zeros(n_sims)
    for j in range(len(start_rpos_array)):
        # Generate a random number between 0 and 1
        rand_num = np.random.uniform(low=0, high=1)
        # Apply the inverse cdf function (done numerically here) to rand_num to obtain the starting radial position
        lhs, rhs = 0, 1
        while rhs <= len(cdf_array) - 1:
            # When found, perform linear intrapolation between data points
            if rand_num >= cdf_array[lhs] and rand_num <= cdf_array[rhs]:
                gradient = (cdf_array[rhs] - cdf_array[lhs])/(pos_r_np[rhs] - pos_r_np[lhs])
                constant = cdf_array[lhs] - gradient*pos_r_np[lhs]
                start_rpos = (rand_num - constant)*1e6/gradient             # in um, for use in track_particle function from PTA_numba.py
                break
            else:
                lhs += 1
                rhs += 1
        start_rpos_array[j] = start_rpos

    # Output selected information
    if extra_info == True:
        return norm_factor_k, cdf_array, pdf_array, pos_r_np, MRCE
    else:
        return cdf_array, MRCE, start_rpos_array, average_velocity
###########################################################################################

### ------------------------------ SIMULATE SEPARATOR ------------------------------ ###
@jit(nopython=True, parallel=True)
def simulate_separator(all_params, starting_radial_positions, seed):
    cap_cnt = 0
    for i in prange(len(starting_radial_positions)):
        start_rpos = starting_radial_positions[i]
        capture_status = track_particle(all_params=all_params, start_r=start_rpos, time_des=0.01, Brownian=True, seed=(i+2)*seed)
        if capture_status == True:
            cap_cnt += 1
    CE = (cap_cnt/len(starting_radial_positions))*100
    return CE
###########################################################################################


### ------------------------------ GENERATE FLOW PROFILE ------------------------------ ###
def generate_flowprofile(all_params, resolution):
    # Initialize particle radius
    rp = all_params[0]                  # in m

    # Initialize all other parameters
    p_drop = all_params[1]              # in Pa
    sep_length = all_params[2]          # in m
    wire_radius = all_params[3]         # in m
    sep_radius = all_params[4]          # in m
    kappa = all_params[7]               # [-]

    # Prepare upper and lower limits of radial position
    rpos_uplim = sep_radius - rp
    rpos_downlim = wire_radius + rp
    pos_r_np = np.linspace(start=rpos_downlim, stop=rpos_uplim, num=resolution)

    # Prepare velocity profile array
    vf_z_profile = np.zeros(len(pos_r_np))

    for i in range(len(pos_r_np)):
        pos_r = pos_r_np[i]
        vf_z_profile[i] = (p_drop*(sep_radius**2)/(4*eta*sep_length)) * (1 - ((pos_r/sep_radius)**2) - ((1 - kappa**2)/math.log(1/kappa)) * math.log(sep_radius/pos_r))

    return pos_r_np, vf_z_profile
###########################################################################################