# This script outputs the statistical performance of the separator

import numpy as np
import combined_functions
from datetime import datetime as dt
import pandas as pd

# Initialize necessary parameters
rp_int = 250            # nm
flowrate = 0.088        # ml/min
time_step = 0.01        # s
n_sims = 10000          # 10000 trajectories
tube_radius = 500/0.9   # um
kappa = 0.9             # dimensionless (0 < kappa < 1, kappa == wire_radius/tube_radius)
sep_length = 500        # mm

# Initialize all parameters and generate array of starting radial positions (set power_status to True for current flow)
all_params = combined_functions.init_params(particle_radius=rp_int, flowrate=flowrate, sep_length=sep_length, tubing_radius=tube_radius, kappa=kappa, rcap_ratio=0.5, power_status=True)  # no randomness
cdf_array, MRCE, start_rpos_array, average_velocity = combined_functions.generate_startrpos(all_params=all_params, resolution=10000, extra_info=False, n_sims=n_sims)              # random in sampling starting positions (seed 42)

# Simulate separator and calculate its performance
begin_time = dt.now()
CE = combined_functions.simulate_separator(all_params=all_params, starting_radial_positions=start_rpos_array, seed=35)                                                             # If performing sensitivity analysis, change seed to index of study parameters
SE = ( ((100 - MRCE) - (100 - CE))/(100 - MRCE) ) * 100

# Report information and results
print(f"\nSimulation of {n_sims} trajectories took {dt.now() - begin_time}.")
print("\tSeparator dimensions: ")
print(f"\t\tLength: {sep_length} mm")
print(f"\t\tWire radius: {round(tube_radius*kappa, ndigits=2)} um")
print(f"\t\tTube radius: {round(tube_radius, ndigits=2)} um\n")

print("\tOperating specifications: ")
print(f"\t\tParticle radius: {rp_int} nm")
print(f"\t\tFlowrate: {flowrate} ml/min")

print(f"\nSeparator captured {round(CE, ndigits=3)} % of the {n_sims} particles fed.")
print(f"Separator had an efficacy of {round(SE, ndigits=3)} %. This means that particles entering this separator are {round(SE, ndigits=3)} % more likely to end up being captured than not.\n")
